--------------------------------------------------------------------------------
-- Copyright  2021								|
-- All Rights Reserved.								|
-- 										|
-- Permission to use, copy, modify, and distribute this software in any way	|
-- is not granted without expressed written agreement. This copyright applies	|
-- to the all code, data and artwork related to this project and all associated	|
-- projects published by the author.						|
--										|
-- More information on copyright is available from sources on-line.		|
-- Use an actual factual one!							|
--------------------------------------------------------------------------------

local _, NS = ...

local L = NS.TEXTS

local upper, gsub = string.upper, string.gsub

local AddonName, ABDB, AddonNameExtended = "AutoBar"
local Scale = 1
local MainScale = 1
local Padding = 0
local ClassList, ClassLookup, Hooked, IsInit, Timer

function NS.SetAutoBarScales()
	if NS.Versions[AddonName][WOW_PROJECT_ID] and NS.Versions[AddonName][WOW_PROJECT_ID].scale then
		local Settings = NS.Versions[AddonName][WOW_PROJECT_ID]
		Scale = Settings.scale
		MainScale = Settings.mainscale
		Padding = Settings.padding
	end
	return Scale, MainScale, Padding
end

local ClassCvt
local Extensions = { -- See also AltNames and FauxMazzle\Modules\Addons.lua
	"",
	"Classic",
	"BCC",
	"Wrath",
	"Cata",
}

local function IsLoaded(deactivate)
	local LoadCheck = AddonName
	for k, v in pairs(Extensions) do
		if C_AddOns.IsAddOnLoaded(LoadCheck..v) then 
			local class = select(2, UnitClass("player"))
			if AutoBarDB2 and AutoBarDB2.classes and AutoBarDB2.classes[class] then
				ABDB = AutoBarDB2
			else
				ABDB = AutoBarDB
			end
			return true
		end
	end
	if deactivate then
		NS:GetSettings("Addons")[AddonName].inactive = true
	else
		NS.DebugPrint("Not Found:", AddonName)
	end
end

local function BuildClassCvt(FauxMazzle)
	ClassCvt = {}
	for k, v in pairs(FauxMazzle.classes) do
		local c = gsub(k, " ", "")
		c = c == 0 and k or c
		ClassCvt[string.upper(c)] = c
	end
end

local function AssignAutoBar(reset)
	if not IsLoaded(true) then
		return
	end
	L = NS.TEXTS
	if not ClassCvt then
		BuildClassCvt(NS:LoadSettingsAutoBar())
	end
	local yos = NS.Versions[AddonName][WOW_PROJECT_ID].mainyos
--[[
	local LoadCheck = AddonName.."Classic"
	if not C_AddOns.IsAddOnLoaded(LoadCheck) then 
		LoadCheck = AddonName
		if not C_AddOns.IsAddOnLoaded(LoadCheck) then 
			NS:GetSettings("Addons")[AddonName].inactive = true
			return 
		end
	end
]]--
	if not ClassList then
		ClassList = {}
		ClassLookup = {}
		for k, v in pairs(ABDB.classes) do 
			ClassList[upper(gsub(k, " ", ""))] = v
			ClassLookup[upper(gsub(k, " ", ""))] = gsub(k, " ", "")
		end
	end
	local Class = UnitClass("player")
	local BarClass = string.gsub(Class, " ", "")
	local ClassBar = _G["AutoBarClassBar"..BarClass.."Driver"]
	if ClassBar then
		NS.SetAutoBarScales()
		ClassBar:SetScale(Scale)
		ClassBar:ClearAllPoints()
		ClassBar:SetPoint("BOTTOMLEFT", DominosFrame1, "TOPLEFT", 0, -2)
--		ClassBar.class:PositionSave()
		ClassBar.SetPoint = function() end
		local es = ClassBar:GetEffectiveScale()
		local posX = ClassBar:GetLeft() * es
		local posY = ClassBar:GetBottom() * es
--		local Settings = ABDB.classes[Class].barList["AutoBarClassBar"..BarClass]
		for i=1, #CLASS_SORT_ORDER do 
			local Settings = ClassList[CLASS_SORT_ORDER[i]]
			if Settings then 
				local SaveSettings = Settings
				Settings = Settings.barList["AutoBarClassBar"..ClassLookup[CLASS_SORT_ORDER[i]]]
				if not Settings then
					if ClassCvt[ClassLookup[CLASS_SORT_ORDER[i]]] then
						Settings = SaveSettings.barList["AutoBarClassBar"..ClassCvt[ClassLookup[CLASS_SORT_ORDER[i]]]]
					end
				end
				if Settings then
					Settings.hide = false
					Settings.posX = posX
					Settings.posY = posY
					Settings.scale = Scale
					Settings.padding = Padding
				end
			end
		end
	end
	if IsInit or reset then
		IsInit = false
		local StartTime = reset and 0.2 or 2
		C_Timer.After(StartTime, function()
			local ABasic = AutoBarClassBarBasicDriver --_G["AutoBarClassBarBasicDriver"]
			if not ABasic then
				print(L.BadAutoBar)
				return
			end
			local Trink = NS:GetFrame("Utility", "Trinket2")
			ABasic:SetScale(MainScale)
			ABasic:ClearAllPoints()
			ABasic:SetPoint("BOTTOMLEFT", Trink, "BOTTOMRIGHT", 0, yos)-- -4)
			ABasic.FMSetPoint = ABasic.SetPoint
			ABasic.SetPoint = function() end
			ABasic:Show()
			local left = ABasic:GetLeft()
			local bottom = ABasic:GetBottom()
			local scale = ABasic:GetEffectiveScale()
			if left and bottom then
				ABDB.account.barList.AutoBarClassBarBasic.posX = left * scale
				ABDB.account.barList.AutoBarClassBarBasic.posY = bottom * scale
				ABDB.account.barList.AutoBarClassBarBasic.scale = MainScale
			end
		end)
		if IsInit and MuffinWhatsNewFrame then
			MuffinWhatsNewFrame:HookScript("OnShow", function(self) 
				self:Hide()
			end)
		end
	end
	LibStub("AceConfigRegistry-3.0"):NotifyChange(AddonName) ----------- Need to see if this does anything positive
end

local function KillKey(table, value, iskey)
	for k, v in pairs(table) do
		if (iskey and k == value) or (not iskey and v == value) then
			table[k] = nil
		end
	end
end
local function ConfigureAutoBar(reset)
	if not IsLoaded() then
		return
	end
	L = NS.TEXTS
--[[
	local LoadCheck = AddonName.."Classic"
	if not C_AddOns.IsAddOnLoaded(LoadCheck) then 
		LoadCheck = AddonName
		if not C_AddOns.IsAddOnLoaded(LoadCheck) then 
			NS.DebugPrint("Not Found:", AddonName)
			return
		end
	end
]]--
	local GetFrame = NS:GetOption("GetFrame")
	local FauxMazzle = NS:LoadSettingsAutoBar()
	if not ClassCvt then
		BuildClassCvt(FauxMazzle)
	end
	if NS:IsValidVersion(NS.WOW_Project.Burning_Crusade_Classic) or NS:IsValidVersion(NS.WOW_Project.WrathClassic) or NS:IsValidVersion(NS.WOW_Project.Cataclysm_Classic) then
		KillKey(FauxMazzle.account.barList.AutoBarClassBarBasic.buttonKeys, "AutoBarButtonCooldownPotionHealth")
		KillKey(FauxMazzle.account.barList.AutoBarClassBarBasic.buttonKeys, "AutoBarButtonCooldownPotionMana")
		KillKey(FauxMazzle.account.buttonList, "AutoBarButtonCooldownPotionHealth", true)
		KillKey(FauxMazzle.account.buttonList, "AutoBarButtonCooldownPotionMana", true)
	end
	NS.SetAutoBarScales()
	if AutoBarClassBarExtrasDriver then
		AutoBarClassBarExtrasDriver:ClearAllPoints()
		AutoBarClassBarExtrasDriver:SetPoint("LEFT", UIParent, "RIGHT", 20, 0)
	end
	local class = select(2, UnitClass("player"))
	if ABDB.classes[class] then
		local Store = {}
		for k, v in pairs(FauxMazzle.classes) do
			local nc = gsub(k, " ", "")
			nc = string.upper(nc)
			Store[nc] = {}
			NS.Lib:CopyTable(v, Store[nc])
			local _, cb = next(Store[nc].barList)
			cb["allowed_class"] = gsub(string.upper(k), " ", "")
			cb.columns = 12
		end
		table.wipe(FauxMazzle.classes)
		NS.Lib:CopyTable(Store, FauxMazzle.classes)
		FauxMazzle.account.barList.AutoBarClassBarBasic["allowed_class"] = "*"
		FauxMazzle.account.barList.AutoBarClassBarExtras["allowed_class"] = "*"
		FauxMazzle.ldbIcon = nil
	end
	for k,v in pairs(FauxMazzle.classes) do
		if v.barList then
			for _, bar in pairs(v.barList) do
				bar.scale = Scale
			end
		end
	end
	IsInit = true
	ABDB.account.barList = {}
	ABDB.classes = {}
	NS.Lib:CopyTable(FauxMazzle.classes, ABDB.classes, NS.Lib.SaveTables)
	NS.Lib:CopyTable(FauxMazzle.account, ABDB.account, NS.Lib.SaveTables)
	ABDB.ldb_icon = { hide = true, }
end

NS:AddConfig(2, AddonName, AddonName, ConfigureAutoBar, AssignAutoBar, nil, nil)
