--------------------------------------------------------------------------------
-- Copyright  2021								|
-- All Rights Reserved.								|
-- 										|
-- Permission to use, copy, modify, and distribute this software in any way	|
-- is not granted without expressed written agreement. This copyright applies	|
-- to the all code, data and artwork related to this project and all associated	|
-- projects published by the author.						|
--										|
-- More information on copyright is available from sources on-line.		|
-- Use an actual factual one!							|
--------------------------------------------------------------------------------

local _, NS = ...

local AddonName = "DiscordUnitFrames"

local ConfigVersion = 1
-- ConfigFunc shoulkd be replaced with local ConfigFunc(CopyTable, settings)
--	the stuff below
-- end
-- See: FauxMazzle\Modules\AddOns.lua where this is RunScript..ed
local ConfigFunc = {
	"if settings.inactive or not DUF_Settings then return end", -- Not using
	"local Activate, Copy",
	"local Setup = DUF_Settings.FauxMazzle",
	"if not Setup then return end", -- Something wrong so scram
	"Activate = Setup.player.TotemFrame",
	"if settings.totemsonbuffs then",
	"        Copy = Setup.TotemsBuffs",
	"else",
	"        Copy = Setup.TotemsHUD",
	"end",
	"CopyTable(Copy, Activate)",
	"Activate = Setup.player.DEATHKNIGHT",
	"if settings.runetext then",
	"        Activate.cooldownfonthide = false",
	"else",
	"        Activate.cooldownfonthide = true",
	"end",
	"if settings.runecolor then",
	"        Activate.cooldownspeccolors = true",
	"else",
	"        Activate.cooldownspeccolors = false",
	"end",
	"if WOW_PROJECT_ID > WOW_PROJECT_MAINLINE then return end", -- will have to change if game version boss frames started is introduced
	"Activate = Setup.boss",
	"if settings.bossvertical then",
	"        Copy = Setup.BossV",
	"else",
	"        Copy = Setup.BossH",
	"end",
	"CopyTable(Copy, Activate)",
}
NS:DefaultsAddonConfigAdd(AddonName, ConfigVersion, ConfigFunc)

local UnitSizes = {
	player = {
		width = 5, -- ((Unit width + Heath bar with) + HBxoffset) + Unitxoffset (Unitxoffset and HBxoffset are negative)
		height = 5, -- Unit height + (Buff size * 2) + DeBuff size + Buff y + DeBuff y + Buff Y Spacing
	},
	pet = {
		width = 5, -- Portrait width + HB width + MB width (additiona for target text but???)
		height = 5, -- Unit height + (Buff size * 2) + DeBuff size + Buff y + DeBuff y + Buff Y Spacing
	},
	target = {
		width = 5, -- Portrait width + HB Width + MB width + xoffsets
		height = 5, -- Portrait height + Debuff yos + (Debuff height * 2) + Debuff spacing
	},
	party = {
		width = 5, -- Unit width + Target Health bar width
		height = 5, -- Unit height + Buff frame size
	},
	partypet = {
		width = 5, -- HB Width + MB  Width 
		height = 5, -- HB height + Portrait Height + Buff height (+ Name text height - anchored below bars)
	},
	tot = {
		width = 5, -- HB Width + DeBuff size + DeBuff yos (ToT Text (ToT frame identifier sits off the bottom left and extends to the left)
		height = 5, -- HB height + MB Height + DeBuff height + DeBuff yos (see ToT text abore)
	},
}

local ceil, floor = math.floor, math.ceil

local function AssignDUF(reset)
	if not C_AddOns.IsAddOnLoaded(AddonName) then 
		NS:GetSettings("Addons")[AddonName].inactive = true
		return 
	end
	local Settings = NS:GetSettings("Addons")[AddonName]
	-- Default DUF Settings.
	Settings.configs = true -- default indicator for UI setup panel
	Settings.totemsonbuffs = false
	Settings.runetext = true
	Settings.runecolor = true
	Settings.bossvertical = false

--	NS:DefaultsAddonConfigInitialise(AddonName, ConfigFunc, ConfigVersion)
	DUF_Options_LoadSettings("FauxMazzle", nil, true)
end

local function ConfigureDUF(reset)
	if not C_AddOns.IsAddOnLoaded(AddonName) then 
		NS.DebugPrint("Not Found:", AddonName)
		return
	end
--	if not DUF_Settings then
--		C_Timer.After(1, ConfigureDUF)
--		return
--	end
	local FrameOffset = 7
	if NS.Versions[WOW_PROJECT_ID] and NS.Versions[WOW_PROJECT_ID].offset then
		FrameOffset = FrameOffset *  NS.Versions[WOW_PROJECT_ID].offset
	end
	local Scale
	if NS.Versions[AddonName][WOW_PROJECT_ID] and NS.Versions[AddonName][WOW_PROJECT_ID].scale then
		Scale = NS.Versions[AddonName][WOW_PROJECT_ID].scale
	end
	
	local FauxMazzle = NS:LoadSettingsDUF()
	local FauxMazzleBossH = NS:LoadSettingsDUF_BossH()
	local FauxMazzleBossV = NS:LoadSettingsDUF_BossV()
	local TotemsOnBuffs, TotemsOnHUD = NS:LoadSettingsDUF_Totems()
	FauxMazzle.optionscale = 1 -- In case we copied one where we've altered this and not reset.
	if DUF_Settings["FauxMazzle"] then
		table.wipe(DUF_Settings["FauxMazzle"])
	else
		DUF_Settings["FauxMazzle"] = {}
	end
	local Duf = DUF_Settings["FauxMazzle"]
	if Scale then
		local SingleTbl = {
			width=true,
			height=true,
			size=true, 
			xoffset=true,
			yoffset=true,
			durationx=true,
			durationy=true,
			durationheight=true,
			textheight=true,
			spacing=true,
			vspacing=true,
			hspacing=true,
		}
		local MultiTbl = {
			xos=true,
			yos=true,
			bgsize=true,
		}
		local function SetScale(tbl, key, value, stbl, mtbl)
			if stbl[key] then
				tbl[key] = ceil(value * Scale)
				return
			end
			if mtbl[key] and type(value) == "table" then
				for k, v in pairs(value) do
					value[k] = ceil(v * Scale)
				end
			end
		end
		for k, v in pairs(DUF_FRAME_INDICES) do
			NS:AdjustSettings(FauxMazzle[v], SetScale, SingleTbl, MultiTbl)
		end
		NS:AdjustSettings(FauxMazzleBossH, SetScale, SingleTbl, MultiTbl)
		NS:AdjustSettings(FauxMazzleBossV, SetScale, SingleTbl, MultiTbl)
		NS:AdjustSettings(TotemsOnBuffs, SetScale, SingleTbl, MultiTbl)
		NS:AdjustSettings(TotemsOnHUD, SetScale, SingleTbl, MultiTbl)
		local Metrics = NS:GetOption("Metrics")
		local BaseHeight = Metrics.BaseHeight -- Metrics["BaseHeight"]
		local HotSpotSize = Metrics.HotSpotSize -- Metrics["HotSpotSize"]
		local HotSpotX = Metrics.InnerHotSpotX -- Metrics["InnerHotSpotX"]
		local CaveWidth = Metrics.CaveWidth * 0.5
		FauxMazzle.player.yos[1] = BaseHeight + FauxMazzle.player.StatusBar[7].height
		local PartyOuter = (NS:GetFrame("Base", 1):GetWidth() - HotSpotX) + FrameOffset
		
		local Set = FauxMazzle.player
		UnitSizes.player.width = Set.width + Set.StatusBar[1].width + Set.StatusBar[1].xoffset + Set.xos[1] + FrameOffset
		FauxMazzle.pet.xos[1] = UnitSizes.player.width
		FauxMazzle.pet.yos[1] = BaseHeight
		
--		FauxMazzle.party.yos[1] = Metrics["RampartTop"]
		for i= 1, 4 do
			FauxMazzle.party.yos[i] = Metrics.RampartTop -- Metrics["RampartTop"]
		end
		Set = FauxMazzle.party
		UnitSizes.party.width = Set.width + Set.StatusBar[4].width
		FauxMazzle.party.xos[1] = PartyOuter + 5
		FauxMazzle.party.xos[2] = -(CaveWidth + Set.StatusBar[4].width + FrameOffset)
		FauxMazzle.party.xos[3] = CaveWidth + 5
		FauxMazzle.party.xos[4] = -(PartyOuter + Set.StatusBar[4].width + FrameOffset)
		for i= 1, 4 do
			FauxMazzle.partypet.xos[i] = FauxMazzle.party.xos[i]
		end
	end
	if NS:IsValidVersion(NS.WOW_Project.WrathClassic) then
		FauxMazzle.player.DEATHKNIGHT.wrathdefaultrunes = true
		FauxMazzle.player.DEATHKNIGHT.cooldownspeccolors = false
		FauxMazzle.player.DEATHKNIGHT.cooldownfonthide = true
		FauxMazzle.target.yos[1] = 106
		FauxMazzle.target.StatusBar[1].xoffset = -27
		FauxMazzle.target.StatusBar[2].xoffset = 27
	end
	FauxMazzle.BossH = {}
	FauxMazzle.BossV = {}
	FauxMazzle.TotemsBuffs = {}
	FauxMazzle.TotemsHUD = {}
	NS.Lib:CopyTable(FauxMazzleBossH, FauxMazzle.BossH)
	NS.Lib:CopyTable(FauxMazzleBossV, FauxMazzle.BossV)
	NS.Lib:CopyTable(TotemsOnBuffs, FauxMazzle.TotemsBuffs)
	NS.Lib:CopyTable(TotemsOnHUD, FauxMazzle.TotemsHUD)
	NS.Lib:CopyTable(FauxMazzle, Duf)
	-- Replace tabs with 8 spaces. CopyTable and settings are passed as part of RunScript
	NS:DefaultsAddonConfigInitialise(AddonName, ConfigFunc, ConfigVersion)
end
NS:AddConfig(1, AddonName, "Discord Unit Frames", ConfigureDUF, AssignDUF)

local function DUFOptions(parent, addonlistentry)
--NS:DefaultsAddonConfigInitialise(AddonName, ConfigFunc, ConfigVersion)
	if addonlistentry.frame then
		addonlistentry.frame:Show()
		return addonlistentry.frame
	end
	local L = NS.TEXTS
	local Container = CreateFrame("Frame", "DUFOptions", parent)
	NS.AddBackground(Container)
	Container:SetSize(120, parent:GetHeight())

	Container.TotemLabel = NS:CreateLabel(Container, "Totems", L.Totems)
	Container.TotemLabel:SetPoint("TOP", 0, -20)--, Container.DKRuneLabel, "BOTTOM", 0, -70)
	Container.TotemsOnBuffs = NS.Lib:CreateLedCheckButton("TotemsOnBuffs", Container, NS:GetSettings("Addons")[addonlistentry.addon], "totemsonbuffs", "Green")
	NS:AddCheckLabel(Container.TotemsOnBuffs, L.OverBuffs, "RIGHT")
	Container.TotemsOnBuffs:SetPoint("TOPLEFT", 10, -35)
	function Container.TotemsOnBuffs:PostUpdate()
		NS:GetOption("InitialiseAddons")(nil, AddonName)
		DUF_Options_LoadSettings("FauxMazzle", nil, true)
	end

	if NS:IsValidVersion(WOW_PROJECT_MAINLINE) or NS:IsValidVersion(NS.WOW_Project.WrathClassic)  then
		Container.DKRuneLabel = NS:CreateLabel(Container, "DKRune", L.DeathKnightRunes)
		Container.DKRuneLabel:SetPoint("TOP", Container.TotemLabel, "BOTTOM", 0, -60)
		Container.BossLabel = NS:CreateLabel(Container, "Boss", L.BossFrames)
		Container.BossLabel:SetPoint("TOP", Container.DKRuneLabel, "BOTTOM", 0, -85)

		Container.DKRuneTexts = NS.Lib:CreateLedCheckButton("DKRuneTexts", Container, NS:GetSettings("Addons")[addonlistentry.addon], "runetext", "Green")
		NS:AddCheckLabel(Container.DKRuneTexts, L.Texts, "RIGHT")
		Container.DKRuneTexts:SetPoint("TOPLEFT", 5, -100)
		Container.DKRuneColor = NS.Lib:CreateLedCheckButton("DKRuneColor", Container, NS:GetSettings("Addons")[addonlistentry.addon], "runecolor", "Green")
		NS:AddCheckLabel(Container.DKRuneColor, L.SpecColors, "RIGHT")
		Container.DKRuneColor:SetPoint("TOPLEFT", Container.DKRuneTexts, "BOTTOMLEFT", 0, -10)
	end
	if Boss1TargetFrame then
		Container.BossVertical = NS.Lib:CreateLedCheckButton("BossVertical", Container, NS:GetSettings("Addons")[addonlistentry.addon], "bossvertical", "Green")
		NS:AddCheckLabel(Container.BossVertical, L.Vertical, "RIGHT")
		Container.BossVertical:SetPoint("TOPLEFT", Container.DKRuneColor, "BOTTOMLEFT", 0, -40)
	
	end
	return Container
end
NS:RegisterAddonOptions(AddonName, DUFOptions)
