--------------------------------------------------------------------------------
-- Copyright  2021								|
-- All Rights Reserved.								|
-- 										|
-- Permission to use, copy, modify, and distribute this software in any way	|
-- is not granted without expressed written agreement. This copyright applies	|
-- to the all code, data and artwork related to this project and all associated	|
-- projects published by the author.						|
--										|
-- More information on copyright is available from sources on-line.		|
-- Use an actual factual one!							|
--------------------------------------------------------------------------------

local AddonName, NS = ...

local clip, upper = string.trim, string.upper

local L

function NS.BaseMailEntry()
	local MailButton, MailParent, MailChild
	if MiniMapMailFrame then 
		MailButton = "MiniMapMailFrame" 
	elseif MinimapCluster and MinimapCluster.IndicatorFrame and MinimapCluster.IndicatorFrame.MailFrame then
		MailParent = "MinimapCluster"
		MailChild = "IndicatorFrame"
		MailButton = "MailFrame"
	elseif MinimapCluster and MinimapCluster.MailFrame then
		MailParent = "MinimapCluster"
		MailButton = "MailFrame"
	elseif MailFrame then
		MailButton = "MailFrame"
	else
		NS.DebugPrint("No Mail Frame Found!", MailParent, MailButton)
	end

	local MailEntry
	if MailChild then
		MailEntry = { button={ [1]=MailParent, [2]=MailChild, [3]=MailButton, }, offsety=2, system=true, }
	elseif MailParent then
		MailEntry = { button={ [1]=MailParent, [2]=MailButton, }, offsety=2, system=true, }
	elseif MailButton then
		MailEntry = { button=MailButton, offsety=2, system=true, }
	else
		MailEntry = {}
	end
	return MailEntry
end

function NS:DefaultsResetCharacter() -- Per Character Settings
	local addonname = NS:GetOption("GlobalName")
	if _G[addonname.."Char"] then
		table.wipe(_G[addonname.."Char"])
	else
		_G[addonname.."Char"] = {}
	end
	if _G[addonname.."_Character"] then
		_G[addonname.."_Character"] = nil -- kill of FauxMazzle 1 Character SVs
	end
	local Settings = _G[addonname.."Char"]
	Settings[0]="*** Temporary Information Only (Chat history/AFK recording etc.) ***" 
	local className, classFilename, classID = UnitClass("player")
	Settings.CONFIG = {
		UI = { point="TOPLEFT", 
			topoint="TOPLEFT", 
			x=250, 
			y=-120, 
		},
	}
	if C_ClassColor then
		Settings.CONFIG.UI.color = C_ClassColor.GetClassColor(classFilename)
	else 
		Settings.CONFIG.UI.color = RAID_CLASS_COLORS[classFilename]
	end
end

function NS.DefaultsInitialiseChat() -- Must run after PLAYER_ENTERING_WORLD but handed to DiscordLib_RegisterInitFunc for now
	L = L or NS.TEXTS
	local Settings = NS:GetSettings("Chat")--.channelbuttons
	if Settings.channelbuttons then
		table.wipe(Settings.channelbuttons)
	else
		Settings.channelbuttons = {}
	end
	if Settings.playback then
		table.wipe(Settings.playback)
	else
		Settings.playback = {}
	end
	for k, v in pairs(Settings) do
		if k ~= "channelbuttons" and k ~= "playback" then
			v = nil
		end
	end
	local ChatFont, ChatFontSize = NS:GetOption("SharedMediaFont"), NS:GetOption("FontSize") --ChatFrame1:GetFont()
	local EditFontSize = NS:GetOption("FontSize")
	Settings.autoguildchannel = nil
	Settings.chatgeneral = L.General
	Settings.delaychannelupdate = nil
	Settings.lockactive = nil
	Settings.hide = nil -- hide chat channel selection buttons on MouseOver. Default nil/false
	Settings.printmoney = nil
	Settings.socialdisplay = nil
	Settings.voicedisplay = nil
	Settings.chatfont = { font=ChatFont, size=1, }
	Settings.editfont = { size=1, }
	Settings.channelbuttons[1] = { name=L.Guild, id=5, }
	Settings.channelbuttons[2] = { name=L.Party, id=6, }
	Settings.channelbuttons[3] = { name=L.Whisper, id=7, }
	Settings.channelbuttons[4] = { name=L.Trade, id=8, }
	Settings.playback.recording = { enabled=true, }
	Settings.playback.history = { enabled=true, }
	Settings.playback.afk = { enabled=true, }
--	Settings.bubbles???
--	Settings.altpos = true -- Place button bar bottom left of screen Default nil/false

	local ChatNames = {
		[1] = L.General,
		[2] = L.CombatLog,
		[3] = VOICE,
		[4] = L.Loot,
		[5] = L.Guild,
		[6] = L.Party,
		[7] = L.Whisper,
		[8] = L.Trade,
	}

	local Messages = {
		Add = {
--			[3] = {
--				"COMBAT_FACTION_CHANGE",
--				"COMBAT_XP_GAIN",
--				"COMBAT_HONOR_GAIN",
--			},
			[4] = {
				"SKILL",
				"LOOT",
				"MONEY",
				"CURRENCY",
				"OPENING",
				"TRADESKILLS",
				"ACHIEVEMENT",
				"GUILD_ACHIEVEMENT",
			},
			[5] = {
				"GUILD",
				"OFFICER",
				"GUILD_ACHIEVEMENT",
			},
			[6] = {
				"PARTY",
				"PARTY_LEADER",
				"RAID",
				"RAID_LEADER",
				"RAID_WARNING",
				"INSTANCE_CHAT",
				"INSTANCE_CHAT_LEADER",
				"BG_HORDE",
				"BG_ALLIANCE",
				"BG_NEUTRAL",
				"COMBAT_FACTION_CHANGE",
			},
			[7] = {
				"WHISPER",
				"BN_WHISPER",
				"BN_INLINE_TOAST_ALERT",
--				"BN_CONVERSATION",
				"BN_WHISPER_PLAYER_OFFLINE",
				"AFK",
				"DND",
			},
			[8] = {
				"CHANNEL",
			},
		},
		Remove = {
			[1] = {
				"SKILL",
				"LOOT",
				"MONEY",
				"CURRENCY",
				"OPENING",
				"TRADESKILLS",
				"ACHIEVEMENT",
				"GUILD_ACHIEVEMENT",
			},
		},
	}
	
	local Channels = {
		Add = {
			[8] = {
				{ 
--					num=2, 
					name="Trade", 
				},
			},
		},
		Remove = {
			[1] = {
				{ 
--					num=2, 
					name="Trade", 
				},
			},
		},
	}

	local SaveSettings = { }
	for i=9, NUM_CHAT_WINDOWS do
		NS:SaveChatFrameSettings(i, SaveSettings)
		SaveSettings[i].Info[7] = false -- shown = false
	end
	local RunCleanup = true
	for i=1, NUM_CHAT_WINDOWS do
		local chatFrame = _G["ChatFrame"..i]
		for index, value in pairs(chatFrame.messageTypeList) do
			if ChatTypeGroup[value] == nil then
				-- Probably BN_WHISPER_PLAYER_OFFLINE and maybe others
				RunCleanup = false
				break
			end
		end
	end
	if RunCleanup then
		FCF_ResetChatWindows()
	end
	if NS:IsValidVersion(WOW_PROJECT_MAINLINE) then
		SetCVar("whisperMode", "popout_and_inline")
	end
	for id=1, #ChatNames do
		local f = _G["ChatFrame"..id]
		SetChatWindowDocked(id, nil)
--		if id == 2 then
--			SetChatWindowUninteractable(id, true)
--			SetChatWindowLocked(id, false)
--		elseif id == 3 or id == 4 then
		if id == 2 or id == 4 then
			if id == 4 then
				SetChatWindowUninteractable(id, false)
			else
				SetChatWindowUninteractable(id, true)
			end
			SetChatWindowLocked(id, true)
		elseif id == 3 then
			SetChatWindowUninteractable(id, true)
			SetChatWindowShown(id, false)
		elseif id < 9 then
			SetChatWindowUninteractable(id, false)
			SetChatWindowLocked(id, true)
		else
			SetChatWindowUninteractable(id, true)
			SetChatWindowLocked(id, false)
		end
		if ChatNames[id] then
			NS:SetChatWindowName(id, ChatNames[id])
			SetChatWindowShown(id, true)
		end
		if Messages.Add[id] then
			for i=1, #Messages.Add[id] do
				local Group = Messages.Add[id][i] --"GUILD_ACHIEVEMENT"
				local Temp
				if not ChatTypeGroup[Group] then
					Temp = true
					ChatTypeGroup[Group] = {}
					if Group == "GUILD_ACHIEVEMENT" then
						for k, v in pairs(CHAT_CATEGORY_LIST.GUILD) do
							tinsert(ChatTypeGroup[Group], "CHAT_MSG_" .. v)
						end
					else
						tinsert(ChatTypeGroup[Group], "CHAT_MSG_" .. Group)
					end
				
				end
				ChatFrame_AddMessageGroup(f, Group)
				if Temp then
					ChatTypeGroup[Group] = nil
				end
			end
		end
		if Messages.Remove[id] then
			for i=1, #Messages.Remove[id] do
				local Group = Messages.Remove[id][i] --"GUILD_ACHIEVEMENT"
				ChatFrame_RemoveMessageGroup(f, Group)
				if Group == "GUILD_ACHIEVEMENT" then
					for k, v in pairs(CHAT_CATEGORY_LIST.GUILD) do
						ChatFrame_RemoveMessageGroup(f, v)
						for index, value in pairs(f.messageTypeList) do
							if upper(value) == upper(v) then
								f.messageTypeList[index] = nil
							end
						end
						f:UnregisterEvent("CHAT_MSG_" .. v)
						RemoveChatWindowMessages(id, v)
					end
				end
			end
		end
		if Channels.Add[id] then 
			for i=1, #Channels.Add[id] do
--				if not NS:ChannelUpdateDelayed(id, "add", Channels.Add[id][i].num, Channels.Add[id][i].name, GetChannelList()) then
				if not NS:ChannelUpdateDelayed(id, "add", Channels.Add[id][i].name, GetChannelList()) then
					ChatFrame_AddChannel(f, Channels.Add[id][i].name)
				end
			end
		end
		if Channels.Remove[id] then
			for i=1, #Channels.Remove[id] do
--				if not NS:ChannelUpdateDelayed(id, "remove", Channels.Remove[id][i].num, Channels.Remove[id][i].name, GetChannelList()) then
				if not NS:ChannelUpdateDelayed(id, "remove", Channels.Remove[id][i].name, GetChannelList()) then
					ChatFrame_RemoveChannel(f, Channels.Remove[id][i].name)
				end
			end
			
		end
	end
	NS:RetstoreChatFrameSettings(SaveSettings)
	NS:GetOption("SetChatChannelsDelayed")()
end

function NS:DefaultsInitialiseCharacter(profile)
	SetCVar("chatBubbles", "0")
	SetCVar("chatBubblesParty", "0")
	L = NS.TEXTS
	local LastVersion
	local Settings = NS:GetSettings(nil, "Profiles")
	if Settings[profile] then
		wipe(Settings[profile])
	else
		Settings[profile] = {}
	end
	local Profile = Settings[profile]
	-- all caps for primary keys in profiles, enfored in GetSettings

	local CharacterDefaults = {
		CHAT = {
		},
		MENU = { -- Context menu
			page1 = true,
			enabled = true,
			labelcolor = { [1]=1, [2]=1, [3]=0, [4]=1 },
			pages = { {}, {} },
		},
		UI = {
			confirmdelete=true, -- default should be true
			buttonicon=0,
			goldonly=nil,
			hotspots = {},
			LeftButton=1,
			RightButton=2,
			skin = { 
				skin = NS:GetOption("GetDefaultAccessory")("Skin"), 
				ignore = {}, 
				color = true,
				colorcombat={ r=0.717, g=0, b=0.047, a=1,},
				colornotcombat={ r=1, g=1, b=1, a=1, },
			},
			buttontrays = {
				[1] = {
					right = {
						{ button="GameTimeFrame", system=true, },
					},
					left = {
						NS.BaseMailEntry(),
					},
				},
			},
		},
--[[ Maps are somehow different in Classic so let's wait and see
		MAP = {
			enabled = false, -- Not part of old FM so initialise as disabled
			anchor = "TOPLEFT",
			blackout = true,
--			minorscale1 = 0.68, 
--			minorscale2 = 0.85, 
			maxed = 1,
			clicks = {
				[1] = 1, 
				[2] = 1, 
				[3] = 1,
			},
		},
]]--
--[[
		OPTIONS = {
			layoutcolor = NS.LayoutColor,
			layoutalpha = NS.LayoutAlpha,
		},
]]--
		ADDONS = {
			Meters={
				show = 0, -- 0=none, 1=in instance, 2=incombat, 3=both
			},
			RaidFrames1={},
		},
		LOD = {},
		RAIDFRAMES1 = {}
	}

--	Add/Scale/Size minibuttons in tray 1
	local TrackingButton = MiniMapTrackingFrame or MiniMapTrackingFrame or MiniMapTracking or MinimapCluster.Tracking or MinimapCluster.TrackingFrame
	TrackingButton = TrackingButton and TrackingButton:GetName()
	if WOW_PROJECT_ID == WOW_PROJECT_WRATH_CLASSIC or WOW_PROJECT_ID == WOW_PROJECT_CATACLYSM_CLASSIC then
		CharacterDefaults.UI.buttontrays[1].right[1].scale = 0.56 -- GameTimeFrame
		CharacterDefaults.UI.buttontrays[1].right[1].offsety = 1 -- GameTimeFrame
		CharacterDefaults.UI.buttontrays[1].left[1].scale = 0.66 -- "MiniMapMailFrame"
--		local TrackingButton = MiniMapTrackingFrame and "MiniMapTrackingFrame" or "MiniMapTracking"
		tinsert(CharacterDefaults.UI.buttontrays[1].right, { button=TrackingButton, scale=0.67, offsety=-1, system=true, })
		if CharacterDefaults.ADDONS.DiscordUnitFrames then
			CharacterDefaults.ADDONS.DiscordUnitFrames.runetext = false
			CharacterDefaults.ADDONS.DiscordUnitFrames.runecolor = false
		end
	end
	if NS:IsValidVersion(NS.WOW_Project.Classic) then -- [TBC is finished] or NS:IsValidVersion(NS.WOW_Project.Burning_Crusade_Classic, NS.WOW_Project.Burning_Crusade_Classic) then
		CharacterDefaults.UI.buttontrays[1].right[1].scale = 0.41 -- GameTimeFrame
		CharacterDefaults.UI.buttontrays[1].left[1].scale = 0.66 -- "MiniMapMailFrame"
--		local TrackingButton = MiniMapTrackingFrame and "MiniMapTrackingFrame" or "MiniMapTracking"
		tinsert(CharacterDefaults.UI.buttontrays[1].right, { button=TrackingButton, scale=0.50, system=true, })
	end
--	if NS:IsValidVersion(NS.WOW_Project.Classic) then
--		tinsert(CharacterDefaults.UI.buttontrays[1].right, { button="MiniMapTrackingFrame", scale=0.50, system=true, })
--	end
--	if NS:IsValidVersion(NS.WOW_Project.Burning_Crusade_Classic) then
--		tinsert(CharacterDefaults.UI.buttontrays[1].right, { button="MiniMapTracking", system=true, scale=0.50, offsetx=-7, offsety=4, })
--	end
--	if NS:IsValidVersion(WOW_PROJECT_MAINLINE) then -- Removed because MiniMapLFGFrame is in Wrath
		if GetExpansionLevel() > 7 then
--			CharacterDefaults.UI.buttontrays[1].left[1] = { button={ "MinimapCluster", "MailFrame", }, system=true, scale=1.3, offsety=-1, } -- Mailbox Icon
			CharacterDefaults.UI.buttontrays[1].left[1].scale=1.3
			CharacterDefaults.UI.buttontrays[1].left[1].offsety=-1
			CharacterDefaults.UI.buttontrays[1].right[1].scale=1.4 -- GameTimeFrame
			CharacterDefaults.UI.buttontrays[1].right[1].offsety=-4
			CharacterDefaults.UI.buttontrays[1].right[1].offsetx=nil
--			local buttonName = MinimapCluster.Tracking and "Tracking" or nil
--			if not buttonName then
--				buttonName = MinimapCluster.TrackingFrame and "TrackingFrame" or nil
--			end
			if TrackingButton then
				tinsert(CharacterDefaults.UI.buttontrays[1].right, { button={ "MinimapCluster", TrackingButton, }, system=true, scale=1.3, })
			end
			tinsert(CharacterDefaults.UI.buttontrays[1].right, { button="ExpansionLandingPageMinimapButton", system=true, scale=0.49, offsetx=4, offsety=0, })
--		else
			CharacterDefaults.UI.buttontrays[1].right[1].scale=0.78 -- GameTimeFrame
--			tinsert(CharacterDefaults.UI.buttontrays[1].right, { button="MiniMapTracking", system=true, offsetx=-7, offsety=4, })
			tinsert(CharacterDefaults.UI.buttontrays[1].right, { button="GarrisonLandingPageMinimapButton", system=true, scale=0.68, offsetx=4, offsety=-4, })
		end
		local QSB, QSBScale, QSBOffsetx, QSBOffsety = Queuestatusbutton, 0.96, 4, -4-- retail pre-DF
		if not QSB then
			QSB, QSBScale, QSBOffsety = QueueStatusButton, 0.56, -1 -- DF (case changed sheesh)
		end
		if not QSB then
			QSB, QSBScale = QueueStatusMinimapButton, 0.96 -- Classics?
		end
		if not QSB then
			QSB, QSBScale, QSBOffsetx = MiniMapLFGFrame, 0.64, -6 -- Wrath apperently?
			
		end
		if QSB then
			tinsert(CharacterDefaults.UI.buttontrays[1].left, { button=QSB:GetName(), offsetx=QSBOffsetx, offsety=QSBOffsety, scale=QSBScale, system=true, })
		else
--			tinsert(CharacterDefaults.UI.buttontrays[1].left, { button="QueueStatusMinimapButton", offsetx=4, offsety=-4, scale=0.96, system=true, })
			NS.DebugPrint("Not Queueu Status Button Found!")
		end
		local Metrics = NS:GetOption("Metrics")
--[[
		if TalkingHeadFrame then
			CharacterDefaults.LOD.TalkingHeadFrame = { point=NS.TalkingHeadPos[1].value, x=NS.TalkingHeadPos[1].x, y=NS.TalkingHeadPos[1].y }
		end
]]--
--	end
	tinsert(CharacterDefaults.UI.buttontrays[1].right, { button="MiniButton", system=true, stack="Utility", })

	if C_AddOns.IsAddOnLoaded("BugSack") then
		local BugSack = { button="LibDBIcon10_BugSack", scale=0.9, system=true }
		if NS:IsValidVersion(NS.WOW_Project.Classic, NS.WOW_Project.Burning_Crusade_Classic) then -- Check!!
			BugSack.scale = 0.64
		end
		if WOW_PROJECT_ID == NS.WOW_Project.WrathClassic or WOW_PROJECT_ID == NS.WOW_Project.Cataclysm_Classic then
			BugSack.scale = 0.64
			BugSack.offsetx = -4
		end
		tinsert(CharacterDefaults.UI.buttontrays[1].left, BugSack)
			
	end
	local Metrics = NS:GetOption("Metrics")
	local GDefaults = NS:GetSettings(nil, "Defaults").UI
	for i=1, Metrics.MaxUserHotSpots do
		if GDefaults.hotspots[i] then
			CharacterDefaults.UI.hotspots[i] = i
		else
			CharacterDefaults.UI.hotspots[i] = { label="", script="" }
		end
	end
	NS.Lib:CopyTable(CharacterDefaults, Profile)
	NS:DefaultsResetCharacter()
	NS:GetOption("RegisterForInit")(nil, "CHANNEL_UI_UPDATE", "InitialiseChatDefaults", 1, NS.DefaultsInitialiseChat)
end

function NS:DefaultsInitialiseGlobal(reset)
	L = NS.TEXTS
	local addonname = NS:GetOption("GlobalName")
	if FAUXMAZZLE_Options then -- Clear from FM1
		FAUXMAZZLE_Options = nil
	end
	if reset or _G[addonname] then
		table.wipe(_G[addonname])
	else
		_G[addonname] = {}
	end
	_G[addonname].Profiles = {}
	_G[addonname].Defaults = {
		UI = {
			hotspots = {
				[2] = {
					addon="xcalc",
					label=L.HotSpot2,
					script="if not xcalc then return end\nif not Calculator_MainFrame then\n	xcalc.windowdisplay()\n	return\nend\nCalculator_MainFrame:SetShown(not Calculator_MainFrame:IsShown())\n",
				},
				[3] = {
					addon="TinyPad",
					label=L.HotSpot3,
					script="if not TinyPad then return end\nTinyPad.SlashHandler(\"\")",
				},
				[4] = {
					addon="DiscordUnitFrames",
					label=L.HotSpot4.." ("..L.Disabled..")",
					script="if InCombatLockdown() then\n	print(\"" .. L.HotSpot4 .. " " .. L.DisabledInCombat .. "\")\n	return\nend\nDUF_Settings[DUF_INDEX].targettarget.hide = not DUF_Settings[DUF_INDEX].targettarget.hide\nself:SetText(\"" .. L.HotSpot4 .. " (\" .. (DUF_Settings[DUF_INDEX].targettarget.hide and \"" .. L.Disabled .. "\" or \"" .. L.Enabled .. "\") .. \")\")\nDUF_SetFrameUnitWatch(DUF_TargetOfTargetFrame, not DUF_Settings[DUF_INDEX].targettarget.hide) \nDUF_Initialize_Frame(DUF_TargetOfTargetFrame, \"targettarget\")\n",
					onreset="self:SetText(\"" .. L.HotSpot4 .. " (\" .. (DUF_Settings[DUF_INDEX].targettarget.hide and \"" .. L.Disabled .. "\" or \"" .. L.Enabled .. "\") .. \")\")",
				},
				[5] = {
					addon="DiscordUnitFrames",
					label=L.HotSpot5.." ("..L.Disabled..")",
					script="if InCombatLockdown() then\n	print(\"" .. L.HotSpot5 .. " " .. L.DisabledInCombat .. "\")\n	return\nend\nDUF_Settings[DUF_INDEX].focus.hide = not DUF_Settings[DUF_INDEX].focus.hide\nself:SetText(\"" .. L.HotSpot5 .. " (\" .. (DUF_Settings[DUF_INDEX].focus.hide and \"" .. L.Disabled .. "\" or \"" .. L.Enabled .. "\") .. \")\")\nDUF_SetFrameUnitWatch(DUF_FocusFrame, not DUF_Settings[DUF_INDEX].focus.hide)\nDUF_Initialize_Frame(DUF_FocusFrame, \"focus\")\n",
					onreset="self:SetText(\"" .. L.HotSpot5 .. " (\" .. (DUF_Settings[DUF_INDEX].focus.hide and \"" .. L.Disabled .. "\" or \"" .. L.Enabled .. "\") .. \")\")",
				},
				[6] = {
					label=L.HotSpot6,
					script="ToggleWorldMap()",
				},
				[7] = {
					label=L.HotSpot7,
					script="ToggleAllBags()",
				},
				[8] = {
					label=L.HotSpot8,
					script="",
					script="self:ToggleMeterFrame()",
				},
			},
		},
		Addons = {},
	}
	_G[addonname].WOW_PROJECT_ID = WOW_PROJECT_ID
	DiscordLib_RegisterInitFunc(NS.InitialiseAddons)
end

local AddonConfigs = {}
function NS:DefaultsAddonConfigAdd(addon, version, func)
	AddonConfigs[addon] = { version=version, func=func, } 
end

function NS:DefaultsAddonConfigInitialise(addon)
	local addonname = NS:GetOption("GlobalName")
	local ConfigFunc = _G[addonname].Defaults.Addons
	if not ConfigFunc[addon] then
		ConfigFunc[addon] = {}
	end
	local Settings = AddonConfigs[addon]
	if not ConfigFunc[addon].version or ConfigFunc[addon].version ~= Settings.version then
		ConfigFunc[addon].version = Settings.version
		ConfigFunc[addon].configfunc = ""
		for i=1, #Settings.func do
			ConfigFunc[addon].configfunc = ConfigFunc[addon].configfunc..Settings.func[i].."\n"
		end
	end
end

function NS:DefaultsAddonConfigInitialiseAll(optionsframe)
	for k,_ in pairs(AddonConfigs) do
		NS:DefaultsAddonConfigInitialise(k)
	end
	if optionsframe then
		optionsframe:SetReloadHighlight(true)
	end
end
